//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_coar.C,v 1.9 2004/12/15 21:58:24 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_coar.C
//
#include "prom_grid.hh"

#undef __FUNCT__   
#define __FUNCT__ "PromGrid::InitNormal"
/* PromGrid::InitNormal *******************************************
 *
 *  - grid constructor from last (coarsened) grid ~(a copy constructor)
 *  - form data structures of nodes and tets
 *    * nodes lists - w/ adjac (of quads/solids)
 *    * sets nodes data structures
 *  - leave ready for emediate coarsening
 *  - empty 'prometheus_->elemList'
 *
 *  INPUT: 
 *    - lastgrid: finer grid, coarsened
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - perr: PETSc error code (out)
 *  
 */
int PromGrid::InitNormal( PromGrid *lastgrid ) 
{
  int ierr; assert(lastgrid != NULL);
  
  ierr = CoarsePreInit( nLocalNd_, lastgrid );   CHKERRQ(ierr);
  
  if( options_.verbose_>1 && (comm_.mype()==0 || options_.verbose_>2) ) {
    PetscPrintf(PETSC_COMM_SELF,
       "[%d]%s(%d) nNds=%d(loc=%d,ghos=%d),nGlbl=%d,fact=%d -- Coarse\n",
		comm_.mype(), __FUNCT__,lastgrid->getLevel()+1, nNodes_,
		nLocalNd_,nNodes_-nLocalNd_,
		nglobal(), factor_);
  }

  ierr = CoarseNormalInit( lastgrid );   CHKERRQ(ierr);  
  ierr = CoarsePostInit( lastgrid ); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::InitDummy"
/* PromGrid::InitDummy *******************************************
 *
 *  - grid constructor for dummy grid
 *
 *  INPUT: 
 *    - lastgrid: finer grid, coarsened
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code (out)
 */
int PromGrid::InitDummy( PromGrid *lastgrid )
{
  int ierr;
  assert(lastgrid!=NULL); assert( !isActive() && !lastgrid->isActive() );

  ierr = CoarsePreInit( 0, lastgrid ); CHKERRQ(ierr);

  if( (comm_.mype()==1 && options_.verbose_>2) || options_.verbose_>3 ) {
    PetscPrintf(PETSC_COMM_SELF,"[%d]%s(%d) fact=%d -- Dummy Coarse\n",
		comm_.mype(), __FUNCT__,lastgrid->getLevel()+1, factor_ );  
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::InitReduced(splitter)"
/* PromGrid::InitReduced *******************************************
 *
 *  - grid constructor for splitting grids
 *  - 
 *
 *  INPUT: 
 *    - lastgrid: fine grid
 *    - fact: 
 *    - buffersz: size of buff
 *    - buffer: buffer 
 *    - rtag:
 *    - gh_fgid1_flid1_proc1: ghost gid to local on fine or -proc
 *    - ghostlist: real ghosts nodes (emptied)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code (out)
 */
int PromGrid::InitReduced(PromGrid *lastgrid, const int fact,
			  const int buffsz, int *buffer, const int rtag,
			  PromTable<int> &gh_fgid1_flid1_proc1, 
			  PromList<PromMPINode*> &ghostlist )
{
  int ierr, nn = isActive() ? nLocalNd_ : 0;  
  assert( lastgrid!=NULL && lastgrid->isActive() );
  ierr = CoarsePreInit( nn, lastgrid );  CHKERRQ(ierr);

  if( options_.verbose_>1 && (comm_.mype()==0 || options_.verbose_>2) )
    PetscPrintf(PETSC_COMM_WORLD,
		"[%d]%s(%d) nNs=%d(loc=%d,ghos=%d),nGlbl=%d,fact=%d--Coarse\n",
		comm_.mype(),__FUNCT__,lastgrid->getLevel()+1,nNodes_,
		nLocalNd_,nNodes_-nLocalNd_,
		nglobal(),factor_);

  ierr = CoarseReduceInit( lastgrid, buffsz, buffer, rtag,
			   gh_fgid1_flid1_proc1, ghostlist );
  CHKERRQ(ierr);

  ierr = CoarsePostInit( lastgrid ); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::CoarsePreInit"
/* PromGrid::CoarsePreInit ***************************************************
 *
 *  - 
 *
 *  INPUT: 
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::CoarsePreInit( int nloc, PromGrid *lastgrid )
{
  lastgrid->nextgrid_ = this;
  prevgrid_ = lastgrid;

  blockISs_ = NULL;
  nBlocks_ = lastgrid->nBlocks_; // flag for now 

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGrid::CoarsePostInit"
/* PromGrid::CoarsePostInit **************************************************
 *
 *  - 
 *
 *  INPUT: 
 *    - lastgrid:
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::CoarsePostInit( PromGrid *lastgrid )
{
  int ierr;
  
  // setup ghost
  if(comm_.mype()==0 && options_.verbose_>2)
    PetscPrintf(gComm_,"[%d]%s(%d) call ResolveGhostData (factor=%d)\n",
		comm_.mype(),__FUNCT__,lastgrid->getLevel()+1,factor_ );
  
  assert(proc_geq_ != NULL); 
  if( lastgrid->isActive() ) {
    ierr = ResolveGhostData( lastgrid );            CHKERRQ(ierr);
  }
  else assert( proc_geq_[comm_.mype()+1] - proc_geq_[comm_.mype()] == 0 );
  assert(ghosts_.getCount() == nNodes_-nLocalNd_);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::CoarseNormalInit"
/* PromGrid::CoarseNormalInit ***********************************************
 *
 *  - "normal" coarse grid, take fine "coarsened" grid, ready for triang.
 *
 *  INPUT: 
 *    - lastG: finer grid, coarsened
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::CoarseNormalInit( PromGrid *lastG )
{
  const int mype=comm_.mype(),my0=proc_gnode_[mype];
  int ierr,xx,pid,gid,geq;
  PromNode *curr, *fine;
 
  // COPY NODES up, set parent/child, set pid, gid, send dep<locid, pid, gid>
  if( nodes_ == NULL ) return 0; // when not reduced after a reduction 
  for(xx = pid = 0, geq = proc_geq_[mype] ; xx < lastG->nNodes_ ; xx++ ){
    fine = lastG->nodes_[xx];    
    assert( !fine->isMarked() );
    // skip deleted, not done
    if( fine->getState() != PROM_SELECTED ) continue;
    // copy data
    curr = nodes_[pid]; 
    ierr = curr->LiftCopy( fine ); CHKERRQ(ierr);
    curr->mark(FALSE); 
    // make fresh
    assert( curr->getState() == fine->getState() );
    //for( ii = 0; ii < PROM_MAXNDF ; ii++ ) curr->setFixed(ii,FALSE);
    curr->index_ = pid; 
    curr->setExterior(FALSE); // all nodes start inside
    curr->setState(PROM_UNKNOWN); 
    if( pid < nLocalNd_ ) {     // local - gid
      assert( !fine->isGhost() );
      gid = pid + my0; // initial gid
      curr->index_.setGlobal( gid ); 
      curr->setGEQ( geq ); geq += curr->getNDF();
    }
    else assert( fine->isGhost() && curr->isGhost() ); // set gid latter
    // set fine->parent = coarse node to next level 
    curr->symbData_->child = fine; assert(!fine->symbData_->parent);
    fine->symbData_->parent = curr;
    pid++;
  }
  assert( pid == nNodes_ );    
  
  // make approximate adjacentcies for coarse nodes
  ierr = MakeInitAdjacs();                              CHKERRQ(ierr);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::CoarseReduceInit"
/* PromGrid::CoarseReduceInit ***********************************************
 *
 *  - "splitter" coarse grid
 *
 *  INPUT: 
 *    - lastG: fine grid
 *    - buffsz: size of buff
 *    - buff: buffer 
 *    - gh_fgid1__flid1_proc1: ghost gid to local on fine or -proc
 *    - ghostlist: real ghosts nodes (emptied)
 */
int PromGrid::CoarseReduceInit( PromGrid *lastG, const int buffsz, int buff[],
				const int rtag, PromTable<int> &gh_fgid1__flid1_proc1,
				PromList<PromMPINode*> &ghostlist )
{
  int ierr; 
  if(!lastG->isActive()) SETERRQ(1,"lastG->isActive()");

  // receive local nodes
  ierr = SetNodes( lastG, buffsz, buff, rtag, gh_fgid1__flid1_proc1,
		   ghostlist ); CHKERRQ(ierr);    

  // make approximate adjacentcies for coarse nodes
  ierr = MakeInitAdjacs();                              CHKERRQ(ierr);

  //communicate approx adjac
  ierr = ExchangeAdjacs(lastG->factor_, lastG->gComm_); CHKERRQ(ierr);

  return 0;
}

#define ISLOCAL(p) (p>=master&&p<endp)
#undef __FUNCT__  
#define __FUNCT__ "PromGrid::SetNodes()"
/* PromGrid::SetNodes **************************************************
 *
 *  - SetNodes, receives messages from STRUCTURE::MakeSplitCoarseGrid
 *
 *  INPUT: 
 *    - lastG: fine grid
 *    - buffsz: size of buff
 *    - buff: buffer 
 *    - rtag: mpi tag to receive
 *    - gh_fgid1__flid1_proc1: ghost gid to local on fine or -proc
 *    - ghostlist: real ghosts nodes (emptied)
 */
int PromGrid::SetNodes( PromGrid *lastG, const int buffsz, int buff[],
			const int rtag, 
			PromTable<int> &gh_fgid1__flid1_proc1, 
			PromList<PromMPINode*> &ghostlist )
{
  int          ierr,xx,pid,proc,*pb,tt,gid,lid,ii,todo,geq,mylid0=-1;
  assert(lastG!=NULL); const int tag = lastG->getNewTag();
  PromNode    *curr,*ghos,*fine;
  TablePos     tpos;                   assert(lastG->isActive());
  MPI_Status   status;                 assert(rtag!=tag);
  PromMPINode  node, *np;              PromMPINode::GetWrite_size(ii);
  const int    master = masterProc(), endp = getEndProc(), mpisz = ii;
  const int    my0 = proc_gnode_[master], lfact = lastG->factor_;
  const int    mype = comm_.mype();
  PromGlobalID *fdep;
  MPI_Comm lcomm = lastG->gComm_;

  // receive local *<mpinode>
  for( proc=master, pid=mylid0=0, gid = my0, geq = proc_geq_[master] ; 
       proc < endp ; proc += lfact ) {
    MPI_Recv(buff,buffsz,MPI_INT,proc/lfact,rtag,lcomm,&status);
    MPI_Get_count( &status, MPI_INT, &tt ); ii = tt; assert(tt%mpisz==0);
    tt = tt / mpisz;
    if( proc < mype ) mylid0 += tt;
    //
    for( xx = 0, pb = buff ; xx < tt ; xx++, pid++, gid++ ) {
      node.Read( &pb ); 
      if( (size_t)pb%sizeof(gfloat) != 0 ) pb++; // make even
      curr = nodes_[pid];                      assert(pid < nLocalNd_);
      curr->Copy( node );  curr->mark(FALSE); // old fine coord, matid, lid
      for( int i = 0; i < PROM_MAXNDF ; i++ ) curr->setFixed(i,FALSE);
      curr->index_.setGlobal( gid ); curr->index_ = pid;
      curr->setGEQ( geq ); geq += curr->getNDF();
      assert( !curr->isGhost() );

    }
    assert(ii==pb-buff);
  }
  if(pid != nLocalNd_)SETERRQ2(1,"pid(%d) != nLocalNd_(%d)",pid,nLocalNd_);

  // lookup for local id of fine gid of ghosts
  PromTable<int> fgid1_clid1( (nNodes_ - nLocalNd_ + 5) * 2 );
  // copy in ghosts
  for( tt = nLocalNd_ ; !ghostlist.isEmpty() ; tt++ ) {
    np = ghostlist.removeHead(); 
    ghos = nodes_[np->index_()];  
    ghos->Copy( *np ); ghos->mark(FALSE); // setsome as ghosts,old fine gid
    // ghost has old gid
    fgid1_clid1.add( tt+1, np->index_.global()+1 );
    delete np;  np = NULL; 
    assert(!ghos->symbData_->dependents);
  }
  assert(tt == nNodes_);

  // get data from local fine grid, send local ghosts data
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<int> *list; 
  PromList<MPI_Request*> smessages;
  PromTable<PromList<int>*> proc_list(57);
  for( xx = 0, gid = my0 + mylid0, lid = mylid0; 
       xx < lastG->nLocalNd_ ; xx++ ){
    fine = lastG->nodes_[xx];
    if( fine->getState() != PROM_SELECTED ) continue;
    curr = nodes_[lid]; assert(curr->index_.global()==gid);
    geq = curr->geq();
    ierr = curr->LiftCopy( fine );  CHKERRQ(ierr);
    curr->mark(FALSE);   curr->setState(PROM_UNKNOWN);
    curr->index_ = lid;  curr->setExterior(FALSE);
    curr->index_.setGlobal( gid ); 
    for( ii = 0; ii < PROM_MAXNDF ; ii++ ) curr->setFixed(ii,FALSE);
    curr->symbData_->child = fine; assert(!fine->symbData_->parent);
    fine->symbData_->parent = curr;
    assert( !curr->isGhost() ); // curr->realID = -9;
    if( fine->isBound() ) { // see if a real bound, and send messages
      Position pos = fine->symbData_->dependents->getHeadPosition();
      assert( pos != NULL );                                 tt = 0;
      while( pos ) {
	fdep = fine->symbData_->dependents->getNext(pos);
	proc = fdep->proc();
	if( !ISLOCAL(proc) ) tt = 1; // a real boundary
	else { 
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<int>(100); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( fdep->local() + 1 ); list->addTail( lid + 1 ); 
	}
      }
      if( tt == 0 ) assert(!curr->symbData_->dependents);
      else { 
	curr->symbData_->dependents = new PromList<PromGlobalID*>(3);
	CHKNEW(curr->symbData_->dependents,1);
      }
    }
    lid++; gid++;
  }
  // send data
  tpos = proc_list.getHeadPosition(); 
  while( tpos ) { 
    list = proc_list.getNext( tpos, &proc ); proc--;
    ii = (list->getCount() + reqsz)*sizeof(int);
    MPI_Request *sreq; ierr = PetscMalloc(ii,&sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)(sreq+1), *pb = buff;
    while( !list->isEmpty() ) { ii = list->removeHead() - 1; *pb++ = ii; }
    // send(tag) *< flocid, clid >
    MPI_Isend(buff, pb-buff, MPI_INT, proc/lfact ,tag, lcomm, sreq );
    delete list;
  } 
  // set pointers to last grid for old ghosts
  PromList<int> lid_gid_list(100); PromList<PromNode*> fine_list(50);
  for( xx = lastG->nLocalNd_, todo = 0 ; xx < lastG->nNodes_ ; xx++ ){
    fine = lastG->nodes_[xx];
    if( fine->getState() != PROM_SELECTED ) continue; 
    // get ghost curr from fine ghost, set "gid" and "lid", "fine"
    if( (tt=gh_fgid1__flid1_proc1.find(fine->index_.global()+1)) != 0 ) {
      gid = fine->index_.global();       assert( !fine->isMarked() || tt < 0 );
      lid = fgid1_clid1.find(gid+1); assert(lid); lid--;
      assert(lid >= nLocalNd_);
      // store
      lid_gid_list.addTail(lid+1); lid_gid_list.addTail(gid+1); 
      fine_list.addTail( fine );
    }
    else todo++;
  }

  // recv ghosts
  while( todo ) { // get local curr from fine ghost
    MPI_Probe( MPI_ANY_SOURCE, tag, lcomm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    proc = status.MPI_SOURCE * lfact;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); 
    int *pb = buff; CHKERRQ(ierr);
    // receive(tag) *< flocid, clid >
    MPI_Recv(buff,tt,MPI_INT,status.MPI_SOURCE,tag,lcomm,&status);
    while( pb-buff < tt ) {
      todo--; assert(todo>=0);
      ii = *pb++; lid = *pb++;  assert(ii < lastG->nNodes_ );
      fine = lastG->nodes_[ii];
      assert( fine->isGhost() && lid >= 0 && lid < nLocalNd_); 
      gid = lid + my0; 
      // store
      lid_gid_list.addTail( lid + 1 ); lid_gid_list.addTail( gid + 1 ); 
      fine_list.addTail( fine );
    }
    assert(pb-buff == tt);       PetscFree(buff);
  }

  // recv ghosts
  while( !fine_list.isEmpty() ) {
    fine = fine_list.removeHead();
    lid = lid_gid_list.removeHead() - 1; gid = lid_gid_list.removeHead() - 1;
    //
    ghos = nodes_[lid];
    ghos->symbData_->child = fine; assert(!fine->symbData_->parent); 
    fine->symbData_->parent = ghos;// connect
    // copy in
    geq = ghos->geq();
    ierr = ghos->LiftCopy( fine );  CHKERRQ(ierr);
    ghos->mark(FALSE); ghos->setState(PROM_UNKNOWN); 
    ghos->index_ = lid; ghos->index_.setGlobal( gid );
    for( ii = 0; ii < PROM_MAXNDF ; ii++ ) ghos->setFixed(ii,FALSE);
    ghos->setExterior(FALSE);
    if( gh_fgid1__flid1_proc1.find(fine->index_.global()+1) == 0 ){
      // get rid of boudary data structs
      if( ghos->symbData_->dependents ) {
	assert( ghos->symbData_->dependents->isEmpty() );
	delete ghos->symbData_->dependents; 
	ghos->symbData_->dependents = NULL; 
      }
      ghos->symbData_->ghostID = -9;  ghos->symbData_->ghostID.setProc(-9);
      assert(!ghos->isGhost());
    }
    else assert( ghos->isGhost() );
  }
  assert(lid_gid_list.isEmpty());
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE,tag,lcomm,&ii,&status);  assert(!ii);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ExchangeAdjacs"
/* PromGrid::ExchangeAdjacs *******************************************
 *
 *  - splitter procs exchange adjacency data, and faces with sisters
 *
 *  INPUT: 
 *    - lastfactor
 *    - lastcomm
 * 
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::ExchangeAdjacs( const int lastfactor, MPI_Comm lastcomm )
{
  int        ierr,xx,rproc,sproc,*pb,tt,nactive,nlp,ss,nj,mysz,bsz,ii;
  const int  master = masterProc(), endp = getEndProc();
  PromNode  *curr, *n1; const int tag = comm_.getNewTag(lastcomm);
  MPI_Status status;  const int tag1 = comm_.getNewTag(lastcomm);
  MPI_Request sreq;

  // count data
  for( nactive = tt = xx = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];                        assert( curr->symbData_!=NULL);
    if( curr->symbData_->child == NULL ) continue; 
    nactive++;
    tt += curr->symbData_->adjac.getCount() + 2; 
    //ierr = curr->GetWrite_size( ii ); CHKERRQ(ierr); tt += ii;
    //if( tt%gisz != 0 ) tt++;
  }
  mysz = tt; // local size
  for(sproc=master, bsz=mysz, nlp=0; sproc < endp ; sproc += lastfactor,nlp++){
    if( sproc != comm_.mype() ) {
      MPI_Send( &mysz, 1, MPI_INT, sproc/lastfactor, tag, lastcomm );
      MPI_Recv( &tt, 1, MPI_INT, sproc/lastfactor, tag, lastcomm, &status );
      if( tt > bsz ) bsz = tt; // max
    }
  }
  assert(bsz>0);
  // make data to send
  int *sbuff; ierr = PetscMalloc((bsz+1)*sizeof(int),&sbuff); CHKERRQ(ierr);
  int *rbuff; ierr = PetscMalloc((bsz+1)*sizeof(int),&rbuff); CHKERRQ(ierr);
  pb = sbuff; sbuff[bsz]=rbuff[bsz]=-6; assert((size_t)pb%sizeof(gfloat)==0);
  for( xx = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];     assert(curr->getState() == PROM_UNKNOWN);
    if( !curr->symbData_->child ) continue; 
    // send() nnodes * < lid, nadj, <node>, nadj * <lid>, [-2] > 
    *pb++ = xx; *pb++ = curr->symbData_->adjac.getCount(); 
    //ierr = curr->Write( &pb ); CHKERRQ(ierr); 
    TablePos tpos = curr->symbData_->adjac.getHeadPosition();    
    while( tpos ) {
      PromNode *n1 = curr->symbData_->adjac.getNext( tpos );
      *pb++ = n1->index_(); //assert(n1->index_() == n1-nodes);
    }
    //if( (pb-sbuff)%gisz != 0 ) *pb++ = -2;
    nactive--;
  }
  assert(pb-sbuff == mysz && nactive == 0);
  // send
  sproc = (comm_.mype()+lastfactor); 
  rproc = (comm_.mype()-lastfactor); assert(sproc <= master+factor_);
  if( sproc >= endp ) sproc = master;
  if( rproc < master ) rproc = endp - lastfactor;
  // send()/recv() nnodes * < lid, nadj, <node>, nadj*<lid>, [-2] >
  for( xx = 0, ss = mysz; xx < (nlp-1); xx++ ){
    MPI_Isend( sbuff, ss, MPI_INT, sproc/lastfactor, tag1, lastcomm, &sreq );
    MPI_Recv( rbuff, bsz, MPI_INT, rproc/lastfactor, tag1, lastcomm, &status);
    MPI_Get_count( &status, MPI_INT, &tt ); // next send size
    // read it in
    pb = rbuff; 
    while( pb-rbuff < tt ) {
      ii = *pb++; nj = *pb++; assert(ii >= 0 && ii < nNodes_);
      curr = nodes_[ii]; 
      //curr->Read( &pb );//  
      while( nj-- ) {
	ii = *pb++; assert(ii >= 0 && ii < nNodes_);
	n1 = nodes_[ii]; curr->symbData_->adjac.add(n1); //n1->adjac.add(curr);
      }
      //if( (pb-rbuff)%gisz != 0 ) *pb++ = -2;
    }
    assert(pb-rbuff==tt);
    // ready for next
    MPI_Wait( &sreq, &status );
    // swap buffers
    pb = sbuff; sbuff = rbuff; rbuff = pb;
    ss = tt;
  }  
  //
  assert(sbuff[bsz] == -6 && rbuff[bsz] == -6);
  PetscFree(sbuff);
  PetscFree(rbuff);

  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "PromGrid::AddExtraGhostsParents"
/* PromGrid::AddExtraGhostsParents() ***************************
 *
 *  - add extra ghost parent arrays into 'gid_extraParents_'
 *    + uzawa (CC==NULL) use primal matrix, thus, get regularization extras
 *    + non-uzawa (CC!=NULL) get parents of extra ghosts from CR equations
 *
 *  INPUT: 
 *    - nextg:
 *    - CC: 
 *
 *  SIDE EFFECTS:
 *    - add extra ghost parent arrays into gid_extraParents_
 */
int PromGrid::AddExtraGhostsParents( PromGrid * const nextg,
				     const PromMatrix *const CC /*= NULL*/ )
{
  PromMatrix *AA = stiffness_; const PromMatrix* mat;
  int ierr,proc,col,xx,ncols,sz,tt,ii,todo,nloc,cndf;
  const int fact = factor_, np = AA->map_.comm_.np(), my0 = my0Nd();  
  PromList<MPI_Request*> smessages, rmessages;           MPI_Status status;
  sz = sizeof(shpfloat); tt = sizeof(MPI_Request);  ii=sizeof(int);
  const int disz = sz/ii, reqsz = (tt/sz + ((tt%sz == 0) ? 0 : 1)) * disz;
  PromTable<int> *list; PromTable<PromNode*> *tab;
  PromTable<PromTable<int>*> proc_gids; const int tag = getNewTag();
  PromTable<PromTable<PromNode*>*> proc_pararrs; const int tag1 = getNewTag();
  int adjacs[PROM_MAXNIEGH], sbuf[PROM_BSZ], rbuf[PROM_BSZ];

  // (re)set gid_extraParents_
  if( gid_extraParents_ == NULL ) {
    gid_extraParents_ = new PromTable<PromParentArr_extra*>(nLocalNd_/10 + 17);
  }
  else {
    while( !gid_extraParents_->isEmpty() ){
      PromParentArr_extra *parr = gid_extraParents_->removeHead();
      delete parr; 
    }
  }

  if( !isActive() ) return 0; // now nothing to do, 'comm_' not valid
  MPI_Comm_size(gComm_,&tt); assert(comm_.np()/fact == tt);
  
  // collect requests
  for( ii = 0 ; ii < np/fact ; ii++ ) sbuf[ii] = 0; 
  if( CC == NULL ) { cndf = 1; mat = AA; }
  else {
    assert( AA->map_.isConstNDF() ); // const bsz only!
    cndf = AA->map_.constNDFSize(); mat = CC;
  }
  nloc = mat->map_.nlocal();
  //
  for( xx = 0, todo = 0 ; xx < nloc ; xx++ ) {
    // row of fine matrix
    ierr = mat->GetLocalNodeRow(xx,PROM_MAXNIEGH,ncols,adjacs); CHKERRQ(ierr);
    // loop over all the fine nodes (j) - includes self
    for( col = 0; col < ncols ; col += cndf ) { 
      const int gidj = adjacs[col]/cndf, lid = gidj - my0; 
      if( lid < 0 || lid >= nLocalNd_ ) {
	PromNode *finej = ghosts_.find(gidj+1);
	if( finej == NULL ) {
	  // constriant edges
	  ierr = GetProc( gidj, proc ); CHKERRQ(ierr);
	  if( (list=proc_gids.find(proc+1)) == NULL){
	    list = new PromTable<int>(nLocalNd_/10 + 17); CHKNEW(list,1);
	    proc_gids.add( list, proc+1 );
	  }
	  if( list->add( gidj+1 ) ){ sbuf[proc/fact] = 1;  todo++; }
	}
	else if( finej != NULL ) assert( finej->parents_ != NULL );
      }
    }
  }
  // global! -- requests
  for(ii=0;ii<np/fact;ii++) rbuf[ii] = 1;
  MPI_Reduce_scatter( sbuf, &tt, rbuf, MPI_INT, MPI_SUM, gComm_ );
  const int nrecvs = tt; 

  // isend (tag)
  while( !proc_gids.isEmpty() ) {
    list = proc_gids.removeHead( &proc ); proc--;        assert(proc%fact==0);
    tt = list->getCount();                               MPI_Request *sreq; 
    ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq);     CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      int gidg = list->removeHead() - 1; *pb++ = gidg;
    }
    assert(tt==pb-buff);
    // send(tag) *< gidi > 
    MPI_Isend(buff, pb-buff, MPI_INT, proc/fact, tag, gComm_, sreq);
    delete list;
  }
  // recv (tag), collect sends(tag1)
  for( ii = 0 ; ii < nrecvs ; ii++ ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    proc = status.MPI_SOURCE*fact;
    // get table to send back
    if( (tab=proc_pararrs.find(proc+1)) == NULL ) {
      tab = new PromTable<PromNode*>(nLocalNd_/10 + 17); CHKNEW(tab,1);
      proc_pararrs.add( tab, proc+1 );
    }
    MPI_Get_count( &status, MPI_INT, &tt );
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr); 
    // recv(tag) *<gid>
    MPI_Recv(buff,tt,MPI_INT,status.MPI_SOURCE,tag,gComm_,&status);
    while( tt-- ) {
      int gid = *pb++, lid = gid - my0;
      if( !(lid>=0 && lid < nLocalNd_) ) {
	MPI_Get_count( &status, MPI_INT, &ii );
	PromContext::printArr(stderr,buff,ii,"ERROR recv[%d]");
	SETERRQ2(1,"lid=%d gid=%d",lid,gid); 
      }
      tab->add( nodes_[lid] ); // keep bound nodes for send
    }
    PetscFree( buff );
  }
  // free (tag)
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  // send (tag1)
  while( !proc_pararrs.isEmpty() ) {
    tab = proc_pararrs.removeHead( &proc ); proc--;  assert(proc%fact==0);
    // count length
    TablePos tpos = tab->getHeadPosition(); tt = 0;
    while( tpos ) {
      const PromNode *finei = tab->getNext( tpos ); 
      // manual PromParentArr_extra (ugly)
      ierr = finei->parents_->GetWrite_size( sz ); CHKERRQ(ierr);
      tt += 4 + sz; assert(tt%2 == 0);
    }
    assert(tt%disz==0);
    if( tt > 0 ) {
      // send
      MPI_Request *sreq; 
      ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq); CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      while( !tab->isEmpty() ) {
	const PromNode *finei = tab->removeHead();
	int np = finei->nParents();
	*pb++ = finei->index_.global(); 
	*pb++ = finei->geq();  
	*pb++ = finei->isFixed() ? -np : np; 
	*pb++ = finei->getNDF();
	// <gid, geq, (isfixed ? -1 : 1)*np, ndf, pararray>
	ierr = finei->parents_->Write( &pb ); CHKERRQ(ierr);
	assert( (size_t)pb%sizeof(shpfloat) == 0 );
      }
      if(tt!=pb-buff)SETERRQ2(1,"tt!=pb-buff tt=%d pb-buff=%d",tt,pb-buff);
      // send(tag1) *<gid, geq, (isfixed ? -1 : 1)*np, ndf, pararray>
      MPI_Isend(buff,pb-buff, MPI_INT, proc/fact, tag1,gComm_,sreq);
    }
    delete tab; 
  }
  // recv(tag1) *<gid, geq, (isfixed ? -1 : 1)*np, ndf, pararray>
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag1, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); int *buff, *pb;
    ierr = PetscMalloc(tt*sizeof(int),&buff); pb = buff;  CHKERRQ(ierr); 
    // recv(tag1) *<gid, geq, (isfixed ? -1 : 1)*np, ndf, pararray>
    MPI_Recv(buff,tt,MPI_INT,status.MPI_SOURCE,tag1,gComm_,&status);
    while(pb-buff < tt) {
      int gid = *pb++, geq = *pb++, np = *pb++, ndf = *pb++, isfx; 
      if( np < 0 ) { np = -np; isfx = 1; }
      else isfx = 0; 
      PromParentArr_extra *parJArr;
      parJArr = new PromParentArr_extra(parBank_,np,gid,geq,ndf,isfx);
      ierr = parJArr->PromParentArr::Read( &pb );  CHKERRQ(ierr);
      if( CC == NULL ) {
	int add = gid_extraParents_->add( parJArr, gid+1 ); assert(add); // out
      }
      else {
	int add = gid_extraParents_->add( parJArr, geq+1 ); assert(add); // out
      }
      todo--; assert(todo>=0);
    } 
    assert(tt==pb-buff); PetscFree(buff); 
  }
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE,tag,gComm_,&ii,&status);  assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE,tag1,gComm_,&ii,&status);  assert(!ii);

  return 0;
}


#undef __FUNCT__  
#define __FUNCT__ "PromGrid::RepartitionTop"
/* PromGrid::RepartitionTop ************************************************
 *
 *  - constructor
 *
 *  INPUT:
 *    - wcomm:
 *   
 *  SIDE EFFECTS:
 * 
 *  RETURN: 
 *     - PETSc error code
 */
int PromGrid::RepartitionTop( const PromComm &wcomm )
{
  int ierr,xx,nn,proc,*pb,tt,lim,pid,nloc,nloceq,gidi;
  const int my0_orig = my0Nd(), nloc_orig = nlocal();
  PromNode *curr; MPI_Status  status; PromGrid *lastgrid = prevgrid_;
  int *bufd, __buf[PROM_BSZ]; const int tag = lastgrid->getNewTag();
  bufd = ((size_t)__buf%sizeof(double) == 0) ? __buf : __buf + 1;

  // repartition top grid to one
  assert(factor_ != comm_.np() && isTop() );
  if( lastgrid->isActive() ) {  // collect to one proc
    if(options_.verbose_>2)PetscPrintf(comm_(),"[%d]%s(%d)TOP\n",
					comm_.mype(),__FUNCT__,getLevel());
    const int oldfactor = factor_, wasactive = isActive();
    factor_ = comm_.np(); // changes isActive()
    
    // put on one proc
    nBlocks_ = 0; assert(blockISs_ == NULL);
    if( group_ != MPI_GROUP_NULL ) {
      MPI_Group_free( &group_ );            group_ = MPI_GROUP_NULL;
      MPI_Group_free( &mpi_group_world_ ); mpi_group_world_ = MPI_GROUP_NULL;
      if( gComm_ != MPI_COMM_NULL ) MPI_Comm_free( &gComm_ );
      gComm_ = MPI_COMM_NULL;
    }
    // 
    if( isActive() ){
      assert( comm_.mype() == 0 );
      gComm_ = PETSC_COMM_SELF;     // make entire graph here
      // makes non-ghost
      for(xx=nLocalNd_;xx<nNodes_;xx++)nodes_[xx]->symbData_->ghostID.reset();
      // make new array
      nn = nNodes_;
      ierr = AddNNodePointers( nglobal() - nNodes_ ); CHKERRQ(ierr);
      assert(nNodes_ == nglobal());
      for( xx = nn ; xx < nNodes_ ; xx++ ) nodes_[xx] = newNode(xx);
      ghosts_.removeAll();
      // reset adjacs
      for( xx = 0 ; xx < nLocalNd_ ; xx++ ) { 
	curr = nodes_[xx];
	PromTable<PromNode*> *tab = &curr->symbData_->adjac; nn = 0;
	while( !tab->isEmpty() ) { 
	  PromNode* adj = tab->removeHead(); 
	  bufd[nn++] = adj->index_.global();
	}
	while( nn-- ) tab->add( nodes_[ bufd[nn] ] );
      }
      // recieve locals
      for(proc=oldfactor, pid = nLocalNd_; proc<comm_.np(); proc += oldfactor){
	lim = proc_gnode_[proc+oldfactor] - proc_gnode_[proc]; nn = 0;
	while( nn < lim ) {
	  // recv(tag) *< [-3,] numr_node, nadj, [adjacs] >
	  MPI_Recv( bufd, PROM_BSZ, MPI_INT, proc/lastgrid->factor_, tag,
		    lastgrid->gComm_, &status );
	  MPI_Get_count( &status, MPI_INT, &tt ); pb = bufd; 
	  while( pb-bufd < tt ) {
	    curr = nodes_[pid];  
	    if( (size_t)pb%sizeof(double) != 0 ) pb++;
	    ierr = curr->Read( parBank_, &pb ); CHKERRQ( ierr );
	    curr->symbData_->adjac.removeAll(); int nadj = *pb++;
	    while( nadj-- ){ 
	      int gid = *pb++; curr->symbData_->adjac.add( nodes_[gid] );
	    }
	    assert(curr->index_() == pid);
	    pid++; nn++;
	  }
	  assert(tt == pb-bufd);
	}
	assert(nn == lim);
      }
      nLocalNd_ = pid; assert(pid == nglobal() && nNodes_ == nglobal()); 
      nloc = nglobal(); nloceq = nglobalEq(); 
    } 
    else if( wasactive ){
      gComm_ = MPI_COMM_NULL;  
      // send nodes
      int sz,*pb = bufd; 
      for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
	curr = nodes_[xx]; 	
	PromTable<PromNode*> *adj = &curr->symbData_->adjac; 
	ierr = curr->GetWrite_size( sz );   CHKERRQ(ierr);
	if( (pb-bufd) + sz + 1 + adj->getCount() + 1 > PROM_BSZ ) {
	  // send(tag) *< [-3,] numr_node, nadj, [adjacs] >
	  MPI_Send( bufd, pb-bufd, MPI_INT, 0, tag, lastgrid->gComm_ );
	  pb = bufd; 
	}
	if( (size_t)pb%sizeof(double) != 0 ) *pb++ = -3;
	ierr = curr->Write( &pb );  CHKERRQ(ierr);	
	*pb++ = adj->getCount();
	TablePos tpos = adj->getHeadPosition();
	while( tpos ) {
	  PromNode *n1 = adj->getNext(tpos); 
	  *pb++ = n1->index_.global();
	}
      }
      if( pb-bufd > 0 ) {
	// send(tag) < numr_node, nadj, [adjacs] >
	MPI_Send( bufd, pb-bufd, MPI_INT, 0, tag, lastgrid->gComm_ );
      }
      for(xx=0 ; xx < nNodes_ ; xx++){ierr=Kill(nodes_[xx]); CHKERRQ(ierr);}
      PetscFree( nodes_ ); nodes_ = NULL; 
      nNodes_ = nLocalNd_ = 0;
      nloc = nloceq = 0;
    }
    else {
      factor_ = comm_.np(); assert(gComm_ == MPI_COMM_NULL); nloc = nloceq = 0;
    }
  }
  else{ // real dummy proc
    factor_ = comm_.np(); nloc = nloceq = 0;
  }

  // fix proc_gnode, etc,
  assert(proc_gnode_ != NULL);
  // global and collective
  ierr = ResetMap( nloc, nloceq ); CHKERRQ(ierr);
  if( isActive() ) assert(nloc == nglobal() && nloceq == nglobalEq());
  else assert(0 == nlocal() && 0 == nlocalEq());
  assert(proc_gnode_[0] == 0);
  // reset last parents proc
  for( xx = 0 ; xx < lastgrid->nLocalNd_ ; xx++ ) {
    curr = lastgrid->nodes_[xx];     int jj;
    for( jj = 0 ; jj < curr->nParents() ; jj++ ) { 
      PromParent *par = curr->parents_->data_[jj];
      par->setProc(0);
    }
  }
  // delete and create on one
  int adjacs[PROM_MAXNIEGH], ncol; double *Ai[PROM_MAXNIEGH];
  if( stiffness_ != NULL ) {
    // procs are bad and need more gid_cnd
    ierr = stiffness_->DeleteProctable(); CHKERRQ(ierr); 
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_SLES_SETUP]);
    // kktamg
    assert(prolongator_ != NULL); 
    PromCRMatrix *crp = dynamic_cast<PromCRMatrix*>(prolongator_);
    if( scalarMap_ != NULL ) delete scalarMap_; scalarMap_ = NULL;

    // new stiffness
    assert( !stiffness_->isSymmetricMatrix() );
    PromMatrix *AA, *PP; tt = nloc < 800 ? nloc : 800;
    // factory (yuck)
    if( crp != NULL ) AA = new PromCRMatrix( *this, tt, NULL, 0, NULL );
    else if(!stiffness_->isComplex()) AA=new PromMatrix(*this,tt,NULL,0,NULL);
    else AA = new PromComplexMatrix( *this, tt, NULL, 0, NULL );

    for( xx = 0, gidi = my0_orig; xx < nloc_orig ; xx++, gidi++ ){
      // row of fine matrix -- raw data (including imaginary part)
      ierr = stiffness_->PromMatrix::GetLocalNodeRow(xx,PROM_MAXNIEGH,ncol,
						     adjacs,Ai);
      CHKERRQ(ierr);
      for( int col = 0; col < ncol ; col++ ) {
	int gidj = adjacs[col]; // add it transpose - Ai is trans.
	ierr = AA->PromMatrix::SetValuesBlocked(1,&gidj,1,&gidi,Ai[col],
						INSERT_VALUES);
	CHKERRQ(ierr);
      }
    }
    delete stiffness_; stiffness_ = AA;
    ierr = AA->Assembly(); CHKERRQ(ierr);
    // prolongator
    const PromMap &smap = *lastgrid->scalarMap_; 
    tt = nglobalEq() < 120 ? nglobalEq() : 120; xx = isActive() ? tt : 0;
    // factory (yuck)
    if(crp!=NULL) PP=new PromCRMatrix(smap,xx,NULL,xx,NULL,nloceq,nglobalEq());
    else if(!stiffness_->isComplex()) {
      PP = new PromMatrix(smap,xx,NULL,xx,NULL,nloceq,nglobalEq());
    }
    else PP = new PromComplexMatrix(smap,xx,NULL,xx,NULL,nloceq,nglobalEq());
    // copy P in
    int *ids; double *vals;
    ierr = prolongator_->PromMatrix::GetOwnershipRange( &gidi, &lim ); 
    CHKERRQ(ierr);
    for( /*void*/ ; gidi < lim ; gidi++ ) {
      // row of fine matrix
      ierr = prolongator_->PromMatrix::GetRow( gidi, &ncol, &ids, &vals );
      CHKERRQ(ierr);
      ierr = PP->PromMatrix::SetValues(1,&gidi,ncol,ids,vals,INSERT_VALUES); 
      CHKERRQ( ierr );
      ierr = prolongator_->PromMatrix::RestoreRow( gidi, &ncol, &ids, &vals ); 
      CHKERRQ(ierr);
    }
    delete prolongator_; prolongator_ = PP;
    ierr = PP->Assembly(); CHKERRQ(ierr);

    // kktamg (P)
    perf_mon_.EventEnd( PromContext::prom_logs_[LOG_SLES_SETUP] );
    if( crp != NULL && crp->D_ != NULL ) {
      PromMatrix *cr_prol = crp->D_; crp->D_ = NULL;   
      assert(CRmap_ != NULL); PromMatrix *new_cr_pr;
      int cNcr = cr_prol->getN(), cncr = isActive() ? cNcr : 0;
      const PromCRMatrix *lstA;
      lstA = dynamic_cast<const PromCRMatrix*>(lastgrid->stiffness_);
      const PromMap &flmap_i = lstA->C_->map_;
      new_cr_pr = new PromMatrix(flmap_i, cncr, NULL, 0, NULL, cncr, cNcr);
      crp->D_ = new_cr_pr;
      nloceq = flmap_i.nlocalEq(), pid = flmap_i.my0Eq();
      for( xx = 0 ; xx < nloceq ; xx++, pid++ ) {
	ierr = cr_prol->GetRow(pid, &ncol, &ids, &vals ); CHKERRQ(ierr);
	ierr = new_cr_pr->SetValues(1, &pid, ncol, ids, vals, INSERT_VALUES);
	CHKERRQ(ierr);
	ierr = cr_prol->RestoreRow(pid, &ncol, &ids, &vals ); CHKERRQ(ierr);
      }
      ierr = new_cr_pr->Assembly(); CHKERRQ(ierr);
      delete cr_prol;
      delete CRmap_; CRmap_ = NULL; // recreate this!?!?!?
    }
  }
  else assert(prolongator_ == NULL); 

  return 0;
}
